import { inject } from '@angular/core';
import { upsertStoreEntity, removeStoreEntity } from 'akita-ng-fire';
import { map, tap } from 'rxjs/operators';
import { Database, ref, listVal, stateChanges, ListenEvent, set, push, update, remove } from '@angular/fire/database';

/** Set the configuration for the collection service */
function RealTimeConfig(options = {}) {
    return (constructor) => {
        Object.keys(options).forEach((key) => (constructor[key] = options[key]));
    };
}

class RealTimeService {
    constructor(store, path, rtdb) {
        this.store = store;
        try {
            this.rtdb = rtdb || inject(Database);
        }
        catch (err) {
            throw new Error('RealTimeService requires "Database" provider.');
        }
        this.nodePath = path;
        this.pathRef = ref(this.rtdb, this.path);
    }
    get path() {
        return this.constructor['nodeName'] || this.nodePath;
    }
    get idKey() {
        return this.store ? this.store.idKey : 'id';
    }
    /**
     * Function triggered when adding/updating data to the database
     * @note should be overridden
     */
    formatToDatabase(entity) {
        return entity;
    }
    /**
     * Function triggered when getting data from database
     * @note should be overridden
     */
    formatFromDatabase(entity) {
        return entity;
    }
    /**
     * @description just sync with node you specified when initialized the service class without updating the store
     */
    syncNode() {
        return listVal(this.pathRef)
            .pipe(map(value => this.formatFromDatabase(value)));
    }
    /**
     * @description sync the node with the store. `formatFromDatabase` will be called every time there is data incoming.
     */
    syncNodeWithStore() {
        return stateChanges(this.pathRef).pipe(tap(change => {
            const { snapshot, event } = change;
            switch (event) {
                case ListenEvent.added:
                    upsertStoreEntity(this.store.storeName, this.formatFromDatabase(snapshot.toJSON()), snapshot.key);
                    break;
                case ListenEvent.removed:
                    removeStoreEntity(this.store.storeName, snapshot.key);
                    break;
                case ListenEvent.changed:
                    upsertStoreEntity(this.store.storeName, this.formatFromDatabase(snapshot.toJSON()), snapshot.key);
                    break;
            }
        }));
    }
    /**
     * @description adds a store entity to your database.
     * @param entity to add.
     */
    add(entity) {
        if (entity[this.idKey]) {
            const entityRef = ref(this.rtdb, this.nodePath + '/' + entity[this.idKey]);
            return set(entityRef, this.formatToDatabase(entity));
        }
        if (Array.isArray(entity)) {
            const ids = [];
            const promises = entity.map((e) => {
                const pushRef = push(this.pathRef);
                const id = pushRef.key;
                ids.push(id);
                return set(pushRef, this.formatToDatabase(Object.assign(Object.assign({}, e), { [this.idKey]: id })));
            });
            return Promise.all(promises).then(() => ids);
        }
        else {
            const pushRef = push(this.pathRef);
            const id = pushRef.key;
            return set(pushRef, this.formatToDatabase(Object.assign(Object.assign({}, entity), { [this.idKey]: id })))
                .then(() => id);
        }
    }
    update(idOrEntity, entity) {
        if (Array.isArray(idOrEntity)) {
            return Promise.all(idOrEntity.map((e) => update(this.pathRef, { [e[this.idKey]]: this.formatToDatabase(e) })));
        }
        else {
            if (typeof idOrEntity === 'string') {
                return update(this.pathRef, { [idOrEntity]: this.formatToDatabase(entity) });
            }
            else if (typeof idOrEntity === 'object') {
                const id = idOrEntity[this.idKey];
                return update(this.pathRef, { [id]: this.formatToDatabase(idOrEntity) });
            }
            else {
                return new Error(`Couldn\'t find corresponding entity/ies: ${idOrEntity}, ${entity}`);
            }
        }
    }
    /**
     *
     * @param id of the entity to remove
     */
    remove(id) {
        try {
            const pathRef = ref(this.rtdb, `${this.path}/${id}`);
            return remove(pathRef);
        }
        catch (error) {
            return new Error(`Error while removing entity with this id: ${id}. ${error}`);
        }
    }
    /**
     * @warn clears the node and every child of the node.
     */
    clearNode() {
        return remove(ref(this.rtdb, this.nodePath));
    }
}

/**
 * Generated bundle index. Do not edit.
 */

export { RealTimeConfig, RealTimeService };
//# sourceMappingURL=akita-ng-fire-rtdb.mjs.map
