import { ActiveState, EntityState, EntityStore, getEntityType } from '@datorama/akita';
import { AtomicWrite, PathParams, SyncOptions, UpdateCallback, WriteOptions } from '../utils/types';
import { Observable } from 'rxjs';
import { CollectionReference, DocumentChange, DocumentReference, Firestore, QueryConstraint, Transaction, WriteBatch } from '@angular/fire/firestore';
export declare type CollectionState<E = any> = EntityState<E, string> & ActiveState<string>;
export declare type DocOptions = {
    path: string;
} | {
    id: string;
};
export declare type GetRefs<idOrQuery> = idOrQuery extends (infer I)[] ? DocumentReference[] : idOrQuery extends string ? DocumentReference : CollectionReference;
/** check is an Atomic write is a transaction */
export declare function isTransaction(write: AtomicWrite): write is Transaction;
export declare class CollectionService<S extends EntityState<EntityType, string>, EntityType = getEntityType<S>> {
    protected store?: EntityStore<S>;
    private collectionPath?;
    private idsToListen;
    private memoPath;
    private memoQuery;
    protected db: Firestore;
    /** If true, it will multicast observables from the same ID */
    protected useMemorization: boolean;
    protected includeMetadataChanges: boolean;
    protected onCreate?(entity: EntityType, options: WriteOptions): any;
    protected onUpdate?(entity: Partial<EntityType>, options: WriteOptions): any;
    protected onDelete?(id: string, options: WriteOptions): any;
    createId(): string;
    constructor(store?: EntityStore<S>, collectionPath?: string, db?: Firestore);
    private fromMemo;
    protected getPath(options: PathParams): string;
    get idKey(): string;
    /** The path to the collection in Firestore */
    get path(): string;
    /** A snapshot of the path */
    get currentPath(): string;
    get resetOnUpdate(): boolean;
    get mergeRef(): boolean;
    /**
     * The Angular Fire collection
     * @notice If path is an observable, it becomes an observable.
     */
    get collection(): CollectionReference<EntityType>;
    /**
     * Function triggered when adding/updating data to firestore
     * @note should be overridden
     */
    formatToFirestore(entity: Partial<EntityType>): any;
    /**
     * Function triggered when getting data from firestore
     * @note should be overridden
     */
    formatFromFirestore(entity: any): EntityType;
    /**
     * Sync a store with a collection query of Firestore
     * @param path Path of the collection in Firestore
     * @param queryConstraints Query constraints to filter document of the collection
     * @param syncOptions Options about the store to sync with Firestore
     * @example
     * service.syncCollection({ storeName: 'movies-latest', loading: false }).subscribe();
     */
    syncCollection(syncOptions?: Partial<SyncOptions>): Observable<DocumentChange<EntityType>[]>;
    /**
     * @example
     * service.syncCollection(activePath$).subscribe();
     */
    syncCollection(path: string, syncOptions?: Partial<SyncOptions>): Observable<DocumentChange<EntityType>[]>;
    /**
     * @example
     * service.syncCollection([limit(10)], { storeName: 'movie-latest'}).subscribe();
     */
    syncCollection(queryConstraints: QueryConstraint[], syncOptions?: Partial<SyncOptions>): Observable<DocumentChange<EntityType>[]>;
    /**
     * @example
     * service.syncCollection('movies', [limit(10)], { loading: false }).subscribe();
     */
    syncCollection(path: string, queryConstraints?: QueryConstraint[], syncOptions?: Partial<SyncOptions>): Observable<DocumentChange<EntityType>[]>;
    /**
     * Sync a store with a collection group
     * @param collectionId An id of
     * @param queryConstraints Query constraints to filter document of the collection
     * @param syncOptions Options about the store to sync with Firestore
     */
    syncCollectionGroup(syncOptions?: Partial<SyncOptions>): Observable<DocumentChange<EntityType>[]>;
    syncCollectionGroup(queryConstraints?: QueryConstraint[]): Observable<DocumentChange<EntityType>[]>;
    syncCollectionGroup(collectionId: string, syncOptions?: Partial<SyncOptions>): Observable<DocumentChange<EntityType>[]>;
    syncCollectionGroup(collectionId: string, queryConstraints?: QueryConstraint[], syncOptions?: Partial<SyncOptions>): Observable<DocumentChange<EntityType>[]>;
    /**
     * Sync the store with several documents
     * @param ids An array of ids or an observable of ids
     * @param syncOptions Options on the store to sync to
     */
    syncManyDocs(ids: string[], syncOptions?: Partial<SyncOptions>): any;
    /** @deprecated Use syncManyDocs inside a switchMap instead of passing an observable */
    syncManyDocs(ids$: Observable<string[]>, syncOptions?: Partial<SyncOptions>): any;
    /**
     * Stay in sync with one document
     * @param docOptions An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     * @param syncOptions Options on the store to sync to
     */
    syncDoc(docOptions: DocOptions, syncOptions?: Partial<SyncOptions>): Observable<EntityType>;
    /**
     * Stay in sync with the active entity and set it active in the store
     * @param options A list of ids or An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     * @param syncOptions Options on the store to sync to
     */
    syncActive(options: S['active'] extends any[] ? string[] : DocOptions, syncOptions?: Partial<SyncOptions>): S['active'] extends any[] ? Observable<EntityType[]> : Observable<EntityType>;
    /** Return the reference of the document(s) or collection */
    getRef(options?: Partial<SyncOptions>): CollectionReference;
    getRef(ids?: string[], options?: Partial<SyncOptions>): DocumentReference[];
    getRef(id?: string, options?: Partial<SyncOptions>): DocumentReference;
    /** Return the current value of the path from Firestore */
    getValue(options?: Partial<SyncOptions>): Promise<EntityType[]>;
    getValue(ids?: string[], options?: Partial<SyncOptions>): Promise<EntityType[]>;
    getValue(queryConstraints?: QueryConstraint[], options?: Partial<SyncOptions>): Promise<EntityType[]>;
    getValue(id: string, options?: Partial<SyncOptions>): Promise<EntityType>;
    /** Listen to the change of values of the path from Firestore */
    valueChanges(options?: Partial<PathParams>): Observable<EntityType[]>;
    valueChanges(ids?: string[], options?: Partial<PathParams>): Observable<EntityType[]>;
    valueChanges(queryConstraints?: QueryConstraint[], options?: Partial<PathParams>): Observable<EntityType[]>;
    valueChanges(id: string, options?: Partial<PathParams>): Observable<EntityType>;
    /**
     * Create a batch object.
     * @note alias for `writeBatch(firestore)`
     */
    batch(): WriteBatch;
    /**
     * Run a transaction
     * @note alias for `runTransaction(firestore, cb)`
     */
    runTransaction<T>(cb: (transaction: Transaction) => Promise<T>): Promise<T>;
    /**
     * Create or update entities
     * @param entities One or many entities
     * @param options options to write the document on firestore
     */
    upsert<D extends Partial<EntityType> | Partial<EntityType>[]>(entities: D, options?: WriteOptions): Promise<D extends (infer I)[] ? string[] : string>;
    /**
     * Add a document or a list of document to Firestore
     * @param oneOrMoreEntities A document or a list of document
     * @param options options to write the document on firestore
     */
    add<D extends Partial<EntityType> | Partial<EntityType>[]>(oneOrMoreEntities: D, options?: WriteOptions): Promise<D extends (infer I)[] ? string[] : string>;
    /**
     * Remove one or several document from Firestore
     * @param id A unique or list of id representing the document
     * @param options options to write the document on firestore
     */
    remove(id: string | string[], options?: WriteOptions): Promise<void>;
    /** Remove all document of the collection */
    removeAll(options?: WriteOptions): Promise<void>;
    /**
     * Update one or several document in Firestore
     */
    update(entity: Partial<EntityType> | Partial<EntityType>[], options?: WriteOptions): Promise<void>;
    update(id: string | string[], entityChanges: Partial<EntityType>, options?: WriteOptions): Promise<void>;
    update(ids: string | string[], stateFunction: UpdateCallback<EntityType>, options?: WriteOptions): Promise<Transaction[]>;
}
